# -*- coding: utf-8 -*-
"""Various utility functions for ISA-API."""
from __future__ import absolute_import
import csv
import json
import logging
import os
import re
import sys
import uuid
from functools import reduce
from zipfile import ZipFile
import pandas as pd
import yaml

from isatools import isatab
from isatools.model import (
    Process
)

log = logging.getLogger('isatools')


def format_report_csv(report):
    """Format JSON validation report as CSV string

    :param report: JSON report output from validator
    :return: string representing csv formatted report
    """
    output = ''

    if report['validation_finished']:
        output = 'Validation=success\n'

    for warning in report['warnings']:
        output += str('{},{},{}\n').format(warning['code'], warning['message'],
                                           warning['supplemental'])

    for error in report['errors']:
        output += str('{},{},{}\n').format(error['code'], error['message'],
                                           error['supplemental'])
    return output


def detect_graph_process_pooling(G):
    """Computes whether there is pooling in an experimental graph. This is
    unexpected in some cases so this function is used to flag these cases.

    :param G: DiGraph of the experimental graph
    :return: List of process IDs on which processes are pooling
    """
    report = []

    for node_index in G.nodes():
        node = G.indexes[node_index]
        if isinstance(node, Process) and len(G.in_edges(node_index)) > 1:
            log.info('Possible process pooling detected on: {}'.format(
                ' '.join([node.id, node.executes_protocol.name])
            ))
            report.append(node.id)
    return report


def detect_isatab_process_pooling(fp):
    """Runs the graph process pooling checks on an ISA-Tab

    :param fp: A file-like buffer object pointing to an investigation file
    :return: Report on which files containing process pooling and the
    process IDs inside those tables
    """
    report = []

    ISA = isatab.load(fp)

    for study in ISA.studies:
        log.info('Checking {}'.format(study.filename))
        pooling_list = detect_graph_process_pooling(study.graph)

        if len(pooling_list) > 0:
            report.append({
                study.filename: pooling_list
            })

        for assay in study.assays:
            log.info('Checking {}'.format(assay.filename))
            pooling_list = detect_graph_process_pooling(assay.graph)

            if len(pooling_list) > 0:
                report.append({
                    assay.filename: pooling_list
                })
    return report


def insert_distinct_parameter(table_fp, protocol_ref_to_unpool):
    """A curation function to fix pooling problems

    :param table_fp: A file-like buffer object pointing to a table file
    :param protocol_ref_to_unpool: Reference to which protocol REF column to
    'unpool'
    :return: None
    """
    reader = csv.reader(table_fp, dialect='excel-tab')
    headers = next(reader)  # get column headings
    table_fp.seek(0)

    df = isatab.load_table(table_fp)

    # find protocol ref column by index
    protocol_ref_indices = [x for x, y in enumerate(df.columns) if
                            df[y][0] == protocol_ref_to_unpool]

    if len(protocol_ref_indices) != 1:
        raise IndexError(
            'Could not find Protocol REF with provided value {}'.format(
                protocol_ref_to_unpool))
    distindex = []

    for i in range(0, len(df.index)):
        distindex.append(str(uuid.uuid4())[:8])

    protocol_ref_index = protocol_ref_indices[0]
    name_header = None
    head_from_prot = headers[protocol_ref_index:]

    for x, y in enumerate(head_from_prot):
        if y.endswith(' Name'):
            name_header = y
            break

    if name_header is not None:
        log.debug('Are you sure you want to add a column of hash values in {}? '
              'Y/(N)'.format(name_header))
        confirm = input()
        if confirm == 'Y':
            df[name_header] = distindex
            table_fp.seek(0)
            df.to_csv(table_fp, index=None, header=headers, sep='\t')
    else:
        log.debug('Could not find appropriate column to fill with hashes')


def contains(small_list, big_list):
    """Checks if a list is contained in another list

    :param small_list: The smaller list
    :param big_list: The bigger list in which to find the smaller list
    :return: True if the small list is found inside the big list
    """
    if len(small_list) == 0:
        return False

    for i in iter(range(len(big_list) - len(small_list) + 1)):
        for j in iter(range(len(small_list))):
            if big_list[i + j] != small_list[j]:
                break
        else:
            return i, i + len(small_list)
    return False


def create_isatab_archive(inv_fp, target_filename=None,
                          filter_by_measurement=None):
    """Function to create an ISArchive; option to select by assay
    measurement type

    :param inv_fp: A file-like buffer object pointing to an investigation file
    :param target_filename: Target ZIP file name
    :param filter_by_measurement: Select by measurement type
    :return: List of files zipped if successful, None if not successful
    """
    if target_filename is None:
        target_filename = os.path.join(
            os.path.dirname(inv_fp.name), 'isatab.zip')
    ISA = isatab.load(inv_fp)

    all_files_in_isatab = []
    found_files = []

    for s in ISA.studies:
        if filter_by_measurement is not None:
            log.debug('Selecting ', filter_by_measurement)
            selected_assays = [
                a for a in s.assays if
                a.measurement_type.term == filter_by_measurement]
        else:
            selected_assays = s.assays

        for a in selected_assays:
            all_files_in_isatab += [d.filename for d in a.data_files]
    dirname = os.path.dirname(inv_fp.name)

    for fname in all_files_in_isatab:
        if os.path.isfile(os.path.join(dirname, fname)):
            found_files.append(fname)
    missing_files = [f for f in all_files_in_isatab if f not in found_files]

    if len(missing_files) == 0:
        log.debug('Do zip')
        with ZipFile(target_filename, mode='w') as zip_file:
            # use relative dir_name to avoid absolute path on file names
            zip_file.write(inv_fp.name, arcname=os.path.basename(inv_fp.name))

            for s in ISA.studies:
                zip_file.write(
                    os.path.join(dirname, s.filename), arcname=s.filename)

                for a in selected_assays:
                    zip_file.write(
                        os.path.join(dirname, a.filename), arcname=a.filename)

            for file in all_files_in_isatab:
                zip_file.write(os.path.join(dirname, file), arcname=file)

            log.debug(zip_file.namelist())
            return zip_file.namelist()

    else:
        log.debug('Not zipping')
        log.debug('Missing: ', missing_files)
        return None


def squashstr(string):
    """Squashes a string by removing the spaces and lowering it"""

    nospaces = "".join(string.split())
    return nospaces.lower()


def pyvar(string):
    """Transform a string into a valid Python variable

    :param string: Input string
    :return: Python variable-compatible string
    """
    for ch in string:
        if ch.isalpha() or ch.isdigit():
            pass
        else:
            string = string.replace(ch, '_')
    return string


def recast_columns(columns):
    """Cast columns to Characteristics and Parameter Values

    :param columns: String list of columns to cast
    :return: The casted columns as a list
    """
    casting_map = {
        'Material Type': 'Characteristics[Material Type]',
        'Date': 'Parameter Value[Date]',
        'Performer': 'Parameter Value[Performer]'
    }
    for k, v in casting_map.items():
        columns = list(map(lambda x: v if x == k else x, columns))
    return columns


def pyisatabify(dataframe):
    """Processes a DataFrame into a dict

    :param dataframe: Input DataFrame
    :return: A dict of columns
    """
    columns = dataframe.columns

    pycolumns = []
    col2pymap = {}

    nodecontext = None
    attrcontext = None

    columns = recast_columns(columns=columns)

    for column in columns:
        squashedcol = squashstr(column)

        if squashedcol.endswith(('name', 'file')) or \
                squashedcol == 'protocolref':
            nodecontext = squashedcol
            pycolumns.append(squashedcol)
        elif squashedcol.startswith(
                ('characteristics', 'parametervalue', 'comment',
                 'factorvalue')) and nodecontext is not None:
            attrcontext = squashedcol

            # factor values are not in node context
            if attrcontext == 'factorvalue':
                pycolumns.append(pyvar(attrcontext))
            else:
                pycolumns.append(
                    '{0}__{1}'.format(nodecontext, pyvar(attrcontext)))

        elif squashedcol.startswith(('term', 'unit')) and \
                nodecontext is not None and attrcontext is not None:
            pycolumns.append('{0}__{1}_{2}'.format(nodecontext,
                                                   pyvar(attrcontext),
                                                   pyvar(squashedcol)))

        col2pymap[column] = pycolumns[-1]
    return col2pymap


def factor_query_isatab(df, q):
    """Query a ISA-Tab DataFrame using the Pandas query functions

    :param df: A Pandas DataFrame
    :param q: Query, like "rate is 0.2 and limiting nutrient is sulphur"
    :return: DataFrame sliced on the query
    """
    columns = df.columns
    columns = recast_columns(columns=columns)

    for i, column in enumerate(columns):
        columns[i] = pyvar(column) if \
            column.startswith('Factor Value[') else column

    df.columns = columns

    qlist = q.split(' and ')
    fmt_query = []

    for factor_query in qlist:
        factor_value = factor_query.split(' == ')

        fmt_query_part = "Factor_Value_{0}_ == '{1}'".format(
            pyvar(factor_value[0]), factor_value[1])
        fmt_query.append(fmt_query_part)

    fmt_query = ' and '.join(fmt_query)
    log.debug('running query: {}'.format(fmt_query))
    return df.query(fmt_query)


def check_loadable(tab_dir_root):
    """Checks if each study in a directory full of MetaboLights ISA-Tabs can
    be loaded

    :param tab_dir_root: Root of the MTBLS directories
    :return: None
    """
    for mtbls_dir in [x for x in os.listdir(tab_dir_root) if
                      x.startswith('MTBLS')]:
        try:
            isatab.load(os.path.join(tab_dir_root, mtbls_dir))
            log.debug('{} load OK'.format(mtbls_dir))
        except Exception as e:
            log.debug('{0} load FAIL, reason: {1}'.format(mtbls_dir, e))


def compute_study_factors_on_mtbls(tab_dir_root):
    """Produces study factors report

    :param tab_dir_root: Directory containing MTBLS prefixed ISA-Tab
    directories
    :return: None, output writes to stdout

    Usage:
        >>> compute_study_factors_on_mtbls('tests/data')
        # if tests/data contains tests/data/MTBLS1, tests/data/MTBLS2 etc.

        To write to an output file:
        >>> import sys
        >>> stdout_console = sys.stdout  # save normal stdout
        >>> sys.stdout = open('out.txt')  # set to file
        >>> compute_study_factors_on_mtbls('tests/data')
        >>> sys.stdout = stdout_console  # reset stdout
    """
    for mtbls_dir in [x for x in os.listdir(tab_dir_root)
                      if x.startswith('MTBLS')]:
        study_dir = os.path.join(tab_dir_root, mtbls_dir)
        analyzer = IsaTabAnalyzer(study_dir)
        try:
            analyzer.pprint_study_design_report()
        except ValueError:
            pass


# TODO: is this any useful at all? (by Massi 18/11/2020)
class IsaTabAnalyzer(object):
    """A utility to analyze ISA-Tabs"""

    def __init__(self, path):
        self.path = path

    def generate_study_design_report(self, get_num_study_groups=True,
                                     get_factors=True, get_num_levels=True,
                                     get_levels=True, get_study_groups=True):
        """Generates a study design report
        :return: JSON report
        """
        isa = isatab.load(self.path, skip_load_tables=False)
        study_design_report = []
        raw_data_file_prefix = ('Raw', 'Array', 'Free Induction Decay')
        for study in isa.studies:
            study_key = study.identifier if study.identifier != '' \
                else study.filename
            study_design_report.append({
                'study_key': study_key,
                'total_sources': len(study.sources),
                'total_samples': len(study.samples),
                'assays': []
            })
            with open(os.path.join(self.path, study.filename)) as s_fp:
                s_df = isatab.load_table(s_fp)
                for assay in study.assays:
                    assay_key = '/'.join([assay.filename,
                                          assay.measurement_type.term,
                                          assay.technology_type.term,
                                          assay.technology_platform])
                    assay_report = {
                        'assay_key': assay_key,
                        'num_sources': len(assay.samples),
                        'num_samples': len([x for x in assay.data_files
                                            if x.label.startswith(
                                raw_data_file_prefix)])
                    }
                    with open(os.path.join(self.path, assay.filename)) as a_fp:
                        a_df = isatab.load_table(a_fp)
                        merged_df = pd.merge(s_df, a_df, on='Sample Name')
                        factor_cols = [x for x in merged_df.columns if
                                       x.startswith("Factor Value")]
                        if len(factor_cols) > 0:
                            # add branch to get all if no FVs
                            study_group_factors_df = \
                                merged_df[factor_cols].drop_duplicates()
                            factors_list = [x[13:-1] for x in
                                            study_group_factors_df.columns]
                            queries = []
                            factors_and_levels = {}
                            for i, row in study_group_factors_df.iterrows():
                                fvs = []
                                for x, y in zip(factors_list, row):
                                    fvs.append(' == '.join([x, str(y)]))
                                    try:
                                        factor_and_levels = \
                                            factors_and_levels[x]
                                    except KeyError:
                                        factors_and_levels[x] = set()
                                        factor_and_levels = \
                                            factors_and_levels[x]
                                    factor_and_levels.add(str(y))
                                queries.append(' and '.join(fvs))
                            assay_report['total_study_groups'] = len(queries)
                            assay_report['factors_and_levels'] = []
                            assay_report['group_summary'] = []
                            for k, v in factors_and_levels.items():
                                assay_report['factors_and_levels'].append({
                                    'factor': k,
                                    'num_levels': len(v),
                                })
                            for query in queries:
                                try:
                                    columns = merged_df.columns
                                    columns = recast_columns(columns=columns)
                                    for i, column in enumerate(columns):
                                        columns[i] = pyvar(column) if \
                                            column.startswith(
                                                'Factor Value[') else column
                                    merged_df.columns = columns
                                    qlist = query.split(' and ')
                                    fmt_query = []
                                    for factor_query in qlist:
                                        factor_value = \
                                            factor_query.split(' == ')
                                        fmt_query_part = \
                                            "Factor_Value_{0}_ == '{1}'" \
                                                .format(pyvar(factor_value[0]),
                                                        factor_value[1])
                                        fmt_query.append(fmt_query_part)
                                    fmt_query = ' and '.join(fmt_query)
                                    log.debug('running query: {}'.format(
                                        fmt_query))
                                    df2 = merged_df.query(fmt_query)
                                    data_column = [x for x in merged_df.columns
                                                   if x.startswith(
                                            raw_data_file_prefix)
                                                   and x.endswith(
                                            'Data File')][0]
                                    assay_report['group_summary'].append(
                                        dict(study_group=query,
                                             sources=len(
                                                 list(df2['Source Name']
                                                      .drop_duplicates())),
                                             samples=len(
                                                 list(df2['Sample Name']
                                                      .drop_duplicates())),
                                             raw_files=len(
                                                 list(df2[data_column]
                                                      .drop_duplicates()))
                                             ))
                                except Exception as e:
                                    log.debug("error in query, {}".format(e))
                    study_design_report[-1]['assays'].append(assay_report)
        return study_design_report

    def pprint_study_design_report(self):
        """Pretty prints the study design report"""
        print(json.dumps(self.generate_study_design_report(), indent=4,
                         sort_keys=True))

    def compute_stats(self):
        """Computes some statistics about the ISA-Tab study"""
        isa = isatab.load(self.path, skip_load_tables=False)
        print('-------------------------------------------')
        print('Investigation stats')
        print('-------------------------------------------')
        print('Num ontologies declared: {}'.format(
            len(isa.ontology_source_references)))
        print('Num studies: {}'.format(len(isa.studies)))
        print('Total study assays: {}'.format(
            reduce(lambda x, y: len(x.assays) + len(y.assays), isa.studies)))
        print('Num investigation publications: {}'.format(
            len(isa.publications)))
        print('Total study publications: {}'.format(
            reduce(lambda x, y: len(x.publications) + len(y.publications),
                   isa.studies)))
        print('Num study people: {}'.format(len(isa.contacts)))
        print('Total study people: {}'.format(
            reduce(lambda x, y: len(x.contacts) + len(y.contacts),
                   isa.studies)))
        for study in isa.studies:
            print('-------------------------------------------')
            print('Study stats for {}'.format(study.filename))
            print('-------------------------------------------')
            print('Num assays: {}'.format(len(study.assays)))
            from collections import Counter
            counter = Counter()
            for material in study.sources + study.samples + \
                            study.other_material:
                counter.update(material.characteristics)
            for k, v in counter.items():
                print('{characteristic} used {num} times'.format(
                    characteristic=k, num=v))


def batch_fix_isatabs(settings):
    """Batch fix some ISA-Tabs based on some predefined settings

    :param settings: JSON describing what to fix

    settings = {
        "/path/to/MTBLS1/s_MTBLS1.txt": [
            {
                "factor": "Metabolic syndrome",
                "protocol_ref": None
            },
            {
                "factor": "Gender",
                "protocol_ref": "Sample collection"
            }
        ],
        "/path/to/MTBLS2/s_MTBL2.txt": [
            {
                "factor": "genotype",
                "protocol_ref": None
            }
        ]
    }

    :return: None
    """
    for table_file_path in settings.keys():
        log.debug('Fixing {table_file_path}...'.format(
            table_file_path=table_file_path))
        fixer = IsaTabFixer(table_file_path=table_file_path)
        fixer.fix_factor(
            factor_name=settings[table_file_path]['factor'],
            protocol_ref=settings[table_file_path]['protocol_ref'])


def urlify(s):
    """
    Utility function to remove all problematic characters (whitespaces, etc...)
    :param s: string
    :return: cleaned string
    """
    # Remove all non-word characters (everything except numbers and letters)
    # s = re.sub(r"[^\w\s]", '', s)

    # Replace all runs of whitespace with a single dash
    s = re.sub(r"\s+", '-', s)

    return s


class IsaTabFixer(object):
    """Utilities to fix ISA-Tabs"""

    def __init__(self, table_file_path):
        self.path = table_file_path

    @staticmethod
    def clean_isatab_field_names(field_names):
        """Fixes ISA-Tab header names. Sometimes Pandas adds suffixes to them
        uneccessarily or other junk

        :param field_names: List of field header names
        :return: Fixed field names
        """
        # iterates field names and drops the postfix enums that pandas adds
        for i, field_name in enumerate(field_names):
            if field_name.startswith('Term Source REF'):
                field_names[i] = 'Term Source REF'
            elif field_name.startswith('Term Accession Number'):
                field_names[i] = 'Term Accession Number'
            elif field_name.startswith('Unit'):
                field_names[i] = 'Unit'
            elif 'Characteristics[' in field_name:
                if 'material type' in field_name.lower():
                    field_names[i] = 'Material Type'
                else:
                    try:
                        field_names[i] = field_name[field_name.rindex(
                            '.') + 1:]
                    except ValueError:
                        pass
            elif 'Factor Value[' in field_name:
                try:
                    field_names[i] = field_name[field_name.rindex('.') + 1:]
                except ValueError:
                    pass
            elif 'Parameter Value[' in field_name:
                try:
                    field_names[i] = field_name[field_name.rindex('.') + 1:]
                except ValueError:
                    pass
            elif field_name.endswith('Date'):
                field_names[i] = 'Date'
            elif field_name.endswith('Performer'):
                field_names[i] = 'Performer'
            elif 'Protocol REF' in field_name:
                field_names[i] = 'Protocol REF'
            elif field_name.startswith('Sample Name.'):
                field_names[i] = 'Sample Name'
        return field_names

    def fix_factor(self, factor_name, protocol_ref=None):
        """Fixes a factor if it's supposed to be something else

        :param factor_name: The factor that's incorrect
        :param protocol_ref: The protocol REF to add the Parameter Value
        corresponding to the wrong factor as above
        :return: None
        """
        if protocol_ref is None:
            self.replace_factor_with_source_characteristic(
                factor_name=factor_name)
        else:
            self.replace_factor_with_protocol_parameter_value(
                factor_name=factor_name, protocol_ref=protocol_ref)

    def replace_factor_with_source_characteristic(self, factor_name):
        """Fixes a factor if it's supposed to be a source characteristic
        attached

        :param factor_name: The factor that's incorrect
        :return: None
        """
        table_file_df = isatab.read_tfile(self.path)

        field_names = list(table_file_df.columns)
        clean_field_names = self.clean_isatab_field_names(field_names)

        factor_index = clean_field_names.index(
            'Factor Value[{}]'.format(factor_name))
        source_name_index = clean_field_names.index('Source Name')

        if factor_index < len(field_names) and \
                'Term Source REF' in field_names[factor_index + 1] and \
                'Term Accession' in field_names[factor_index + 2]:
            log.debug(
                'Moving Factor Value[{}] with term columns'.format(
                    factor_name))
            # move Factor Value and Term Source REF and Term Accession columns
            field_names.insert(
                source_name_index + 1,
                field_names[factor_index])
            field_names.insert(
                source_name_index + 2, field_names[factor_index + 1 + 1])
            field_names.insert(
                source_name_index + 3, field_names[factor_index + 2 + 2])

            del field_names[factor_index + 3]  # del Factor Value[{}]
            del field_names[factor_index + 1 + 2]  # del Term Source REF
            del field_names[factor_index + 2 + 1]  # del Term Accession
        elif factor_index < len(field_names) and \
                'Unit' in field_names[factor_index + 1] and \
                'Term Source REF' in field_names[factor_index + 2] and \
                'Term Accession' in field_names[factor_index + 3]:
            log.debug(
                'Moving Factor Value[{}] with unit term columns'.format(
                    factor_name))
            # move Factor Value and Unit as ontology annotation
            field_names.insert(
                source_name_index + 1,
                field_names[factor_index])
            field_names.insert(
                source_name_index + 2, field_names[factor_index + 1 + 1])
            field_names.insert(
                source_name_index + 3, field_names[factor_index + 2 + 2])
            field_names.insert(
                source_name_index + 4, field_names[factor_index + 3 + 3])

            del field_names[factor_index + 4]  # del Factor Value[{}]
            del field_names[factor_index + 1 + 3]  # del Unit
            del field_names[factor_index + 2 + 2]  # del Term Source REF
            del field_names[factor_index + 3 + 1]  # del Term Accession
        elif factor_index < len(field_names) and \
                'Unit' in field_names[factor_index + 1]:
            log.debug(
                'Moving Factor Value[{}] with unit column'.format(factor_name))
            # move Factor Value and Unit columns
            field_names.insert(
                source_name_index + 1,
                field_names[factor_index])
            field_names.insert(
                source_name_index + 2, field_names[factor_index + 1 + 1])

            del field_names[factor_index + 2]  # del Factor Value[{}]
            del field_names[factor_index + 1 + 1]  # del Unit
        else:  # move only the Factor Value column
            log.debug('Moving Factor Value[{}]'.format(factor_name))
            field_names.insert(
                source_name_index + 1,
                field_names[factor_index])
            del field_names[factor_index]  # del Factor Value[{}]

        table_file_df.columns = self.clean_isatab_field_names(field_names)

        # Rename Factor Value column to Characteristics column
        field_names_modified = list(table_file_df.columns)
        field_names_modified[source_name_index + 1] = \
            field_names_modified[source_name_index + 1].replace(
                'Factor Value', 'Characteristics')
        table_file_df.columns = self.clean_isatab_field_names(
            field_names_modified)

        with open(self.path, 'w') as out_fp:
            table_file_df.to_csv(path_or_buf=out_fp, index=False, sep='\t',
                                 encoding='utf-8')

    def replace_factor_with_protocol_parameter_value(
            self, factor_name, protocol_ref):
        """Fixes a factor if it's supposed to be a Parameter Value

        :param factor_name: The factor that's incorrect
        :param protocol_ref: Protocol REF for the new Parameter Value
        :return: None
        """
        table_file_df = isatab.read_tfile(self.path)

        field_names = list(table_file_df.columns)
        clean_field_names = self.clean_isatab_field_names(field_names)

        factor_index = clean_field_names.index(
            'Factor Value[{factor_name}]'.format(factor_name=factor_name))

        with open(self.path) as tfile_fp:
            next(tfile_fp)
            line1 = next(tfile_fp)
            protocol_ref_index = list(
                map(lambda x: x[1:-1] if x[0] == '"' and x[-1] == '"' else x,
                    line1.split('\t'))).index(protocol_ref)

        if protocol_ref_index < 0:
            raise IOError(
                'Could not find protocol ref matching {protocol_ref}'
                    .format(protocol_ref=protocol_ref))

        if factor_index < len(field_names) and \
                'Term Source REF' in field_names[factor_index + 1] and \
                'Term Accession' in field_names[factor_index + 2]:
            log.debug(
                'Moving Factor Value[{}] with term columns'.format(
                    factor_name))
            # move Factor Value and Term Source REF and Term Accession columns
            field_names.insert(
                protocol_ref_index + 1, field_names[factor_index])
            field_names.insert(
                protocol_ref_index + 2, field_names[factor_index + 1 + 1])
            field_names.insert(
                protocol_ref_index + 3, field_names[factor_index + 2 + 2])
            del field_names[factor_index + 3]  # del Factor Value[{}]
            del field_names[factor_index + 1 + 2]  # del Term Source REF
            del field_names[factor_index + 2 + 1]  # del Term Accession
        elif factor_index < len(field_names) and \
                'Unit' in field_names[factor_index + 1] and \
                'Term Source REF' in field_names[factor_index + 2] and \
                'Term Accession' in field_names[factor_index + 3]:
            log.debug(
                'Moving Factor Value[{factor_name}] with unit term columns'
                    .format(factor_name=factor_name))
            # move Factor Value and Unit as ontology annotation
            field_names.insert(
                protocol_ref_index + 1, field_names[factor_index])
            field_names.insert(
                protocol_ref_index + 2, field_names[factor_index + 1 + 1])
            field_names.insert(
                protocol_ref_index + 3, field_names[factor_index + 2 + 2])
            field_names.insert(
                protocol_ref_index + 4, field_names[factor_index + 3 + 3])
            del field_names[factor_index + 4]  # del Factor Value[{}]
            del field_names[factor_index + 1 + 3]  # del Unit
            del field_names[factor_index + 2 + 2]  # del Term Source REF
            del field_names[factor_index + 3 + 1]  # del Term Accession
        elif factor_index < len(field_names) and \
                'Unit' in field_names[factor_index + 1]:
            log.debug(
                'Moving Factor Value[{factor_name}] with unit column'
                    .format(factor_name=factor_name))
            # move Factor Value and Unit columns
            field_names.insert(
                protocol_ref_index + 1, field_names[factor_index])
            field_names.insert(
                protocol_ref_index + 2, field_names[factor_index + 1 + 1])
            del field_names[factor_index + 2]  # del Factor Value[{}]
            del field_names[factor_index + 1 + 1]  # del Unit
        else:  # move only the Factor Value column
            log.debug('Moving Factor Value[{factor_name}]'
                      .format(factor_name=factor_name))
            field_names.insert(
                protocol_ref_index + 1, field_names[factor_index])
            del field_names[factor_index]  # del Factor Value[{}]

        table_file_df.columns = self.clean_isatab_field_names(field_names)

        # Rename Factor Value column to Parameter Value column
        field_names_modified = list(table_file_df.columns)
        field_names_modified[protocol_ref_index + 1] = \
            field_names_modified[protocol_ref_index + 1].replace(
                'Factor Value', 'Parameter Value')
        table_file_df.columns = self.clean_isatab_field_names(
            field_names_modified)

        investigation = isatab.load(
            os.path.dirname(self.path), skip_load_tables=True)
        study = investigation.studies[-1]
        protocol = study.get_prot(protocol_ref)
        if protocol is None:
            raise AttributeError(
                'No protocol with name {protocol_ref} was found'.format(
                    protocol_ref=protocol_ref))
        protocol.add_param(factor_name)
        factor = study.get_factor(factor_name)
        if factor is None:
            raise AttributeError(
                'No factor with name {factor_name} was found'.format(
                    factor_name=factor_name))
        else:
            study.del_factor(name=factor_name, are_you_sure=True)

        study.filename = '{study_filename}.fix'.format(
            study_filename=study.filename)

        isatab.dump(
            investigation, output_path=os.path.dirname(self.path),
            i_file_name='i_Investigation.txt.fix', skip_dump_tables=True)

        with open(os.path.join(
                os.path.dirname(self.path), '{s_filename}.fix'.format(
                    s_filename=os.path.basename(self.path))), 'w') as out_fp:
            table_file_df.to_csv(path_or_buf=out_fp, index=False, sep='\t',
                                 encoding='utf-8')

    def remove_unused_protocols(self):
        """Removes usused protocols

        :return: None
        """
        investigation = isatab.load(os.path.dirname(self.path))
        for study in investigation.studies:
            unused_protocol_names = set(x.name for x in study.protocols)
            for process in study.process_sequence:
                try:
                    unused_protocol_names.remove(
                        process.executes_protocol.name)
                except KeyError:
                    pass
            for assay in study.assays:
                for process in assay.process_sequence:
                    try:
                        unused_protocol_names.remove(
                            process.executes_protocol.name)
                    except KeyError:
                        pass
            log.info('Unused protocols: {}'.format(unused_protocol_names))
            log.info('Location of unused protocols: {}'.format(
                list(map(lambda pr: True if pr.name in unused_protocol_names else False, study.protocols))
            ))
            # remove these protocols from study.protocols
            """
            clean_protocols_list = []
            for protocol in study.protocols:
                if protocol.name not in unused_protocol_names:
                    clean_protocols_list.append(protocol)
            study.protocols = clean_protocols_list
            """
            clean_protocols = [pr for pr in study.protocols if pr.name not in unused_protocol_names]
            log.info('Clean protocol list: {}'.format([pr.name for pr in clean_protocols]))
            study.protocols = clean_protocols
            log.info('Clean study.protocols: {}'.format([pr.name for pr in study.protocols]))
        isatab.dump(
            investigation, output_path=os.path.dirname(self.path),
            i_file_name='{filename}.fix'.format(
                filename=os.path.basename(self.path)), skip_dump_tables=True)


def utf8_text_file_open(path):
    """A function to correctly open utf-8 files in Python 2 and Python 3

    :param path: Path to a file
    :return: A file-like buffer object opened with utf-8 encoding
    """
    if sys.version_info[0] < 3:
        fp = open(path, 'rb')
    else:
        fp = open(path, 'r', newline='', encoding='utf8')
    return fp


def n_digits(num):
    length = 0
    while num / 10 >= 1:
        num //= 10
        length += 1
    return length + 1


# -*- coding: utf-8 -*-
"""Functions for reading and writing SRA-XML.

Functions for reading and writing SRA-XML. SRA-XML content is loaded into
an in-memory representation using the ISA Data Model implemented in the
isatools.model package.
"""
import datetime
import hashlib
import html
import logging
import os
import xml.dom.minidom
from functools import partial

import iso8601
import jinja2
from lxml import etree

from isatools.model import DataFile, OntologyAnnotation, Sample


log = logging.getLogger('isatools')

supported_sra_assays = [
    ('genome sequencing', 'nucleotide sequencing'),
    ('environmental gene survey', 'nucleotide sequencing'),
    ('metagenome sequencing', 'nucleotide sequencing'),
    ('transcription profiling', 'nucleotide sequencing')
]

sra_center_name = 'OXFORD'
sra_broker_name = 'ISAcreator'
sra_lab_name = sra_center_name
sra_submission_action = 'ADD'
sra_center_prj_name = None


def export(investigation, export_path, sra_settings=None, datafilehashes=None):
    """Exports ISA Data model objects to SRA-XML files

    The exporter uses the jinja2 templating engine. The SRA templates can be
    found in isatools/resources/sra_templates

    :param investigation: An Investigation object
    :param export_path: Path to write SRA-XML files to
    :param sra_settings: Some universal settings to apply to the SRA export
    :param datafilehashes: A list of data file hashes to apply to the exported
    files
    :return: None
    """

    def get_comment(assay, name):
        hits = [c for c in assay.comments if c.name.lower() == name.lower()]
        if len(hits) > 1:
            raise AttributeError(
                "Multiple comments of label '{}' found".format(name))
        elif len(hits) < 1:
            return None
        else:
            return hits[0]

    def get_sample(process):
        materials = process.inputs
        sample = None
        for material in materials:
            if isinstance(material, Sample):
                sample = material
                break
        return sample

    def get_pv(process, name):
        hits = [pv for pv in process.parameter_values if
                pv.category.parameter_name.term.lower().replace('_', ' ')
                == name.lower().replace('_', ' ')]
        if len(hits) > 1:
            raise AttributeError(
                "Multiple parameter values of category '{}' found".format(
                    name))
        elif len(hits) < 1:
            return None
        else:
            if isinstance(hits[0].value, OntologyAnnotation):
                value = hits[0].value.term
            else:
                value = hits[0].value
            return value.replace('_', ' ')

    global sra_center_name
    global sra_broker_name
    if sra_settings is not None:
        sra_center_name = sra_settings['sra_center']
        sra_broker_name = sra_settings['sra_broker']
        # sra_lab_name = sra_settings['sra_lab_name']
        # sra_submission_action = sra_settings['sra_submission_action']
        # sra_center_prj_name = sra_settings['sra_center_prj_name']

    log.info('isatools.sra.export()')
    for istudy in investigation.studies:
        is_sra = False
        for iassay in istudy.assays:
            if (iassay.measurement_type.term, iassay.technology_type.term) in \
                    supported_sra_assays:
                is_sra = True
                break
        if not is_sra:
            log.info('No SRA assay found, skipping processing')
            continue

        study_acc = istudy.identifier
        log.debug("sra exporter, working on " + study_acc)

        # Flag SRA contacts for template
        has_sra_contact = False
        for contact in istudy.contacts:
            if "sra inform on status" in \
                    [r.term.lower() for r in contact.roles]:
                contact.inform_on_status = True
                has_sra_contact = True
            if "sra inform on error" in [r.term.lower()
                                         for r in contact.roles]:
                contact.inform_on_error = True
                has_sra_contact = True
        if not has_sra_contact:
            raise ValueError(
                "The study '{0}' has either no SRA contact or no "
                "email specified for the contact. Please ensure you have one "
                "contact with a 'Role' as 'SRA Inform On Status', otherwise "
                "we cannot export to SRA.".format(istudy.identifier)
            )

        if istudy.submission_date is None or istudy.submission_date == '':
            istudy.submission_date = iso8601.parse_date(
                datetime.date.today().isoformat(), iso8601.UTC).isoformat()
        else:
            istudy.submission_date = iso8601.parse_date(
                istudy.submission_date, iso8601.UTC).isoformat()
        istudy.description = html.escape(istudy.description)
        # ideally make it a requirement in the model or JSON to have html
        # escaped content

        env = jinja2.Environment()
        env.loader = jinja2.FileSystemLoader(
            os.path.join(
                os.path.dirname(__file__), 'resources', 'sra_templates'))
        xsub_template = env.get_template('submission_add.xml')
        sra_contact = None
        if sra_settings is not None:
            inform_on_status = sra_settings['sra_broker_inform_on_status']
            inform_on_error = sra_settings['sra_broker_inform_on_error']
            contact_name = sra_settings['sra_broker_contact_name']
            sra_contact = {
                'inform_on_status': inform_on_status,
                'inform_on_error': inform_on_error,
                'contact_name': contact_name
            }
        xsub = xsub_template.render(accession=study_acc,
                                    contacts=istudy.contacts,
                                    submission_date=istudy.submission_date,
                                    sra_center_name=sra_center_name,
                                    sra_broker_name=sra_broker_name,
                                    sra_contact=sra_contact)
        xproj_template = env.get_template('project_set.xml')
        xproj = xproj_template.render(
            study=istudy, sra_center_name=sra_center_name)

        assays_to_export = list()
        for iassay in istudy.assays:
            if (iassay.measurement_type.term, iassay.technology_type.term) in \
                    supported_sra_assays:
                assay_seq_processes = [
                    a for a in iassay.process_sequence if
                    a.executes_protocol.protocol_type.term ==
                    'nucleic acid sequencing']
                for assay_seq_process in assay_seq_processes:
                    do_export = True
                    if get_comment(assay_seq_process, 'export') is not None:
                        log.debug('HAS EXPORT COMMENT IN ASSAY')
                        export = get_comment(assay_seq_process, 'export').value
                        log.debug('export is {}'.format(export))
                        do_export = export.lower() != 'no'
                    else:
                        log.debug('NO EXPORT COMMENT FOUND')
                    log.debug('Perform export? '.format(str(do_export)))
                    if do_export:
                        sample = None
                        curr_process = assay_seq_process
                        while sample is None:
                            sample = get_sample(curr_process)
                            curr_process = curr_process.prev_process
                        assay_to_export = \
                            {
                                'sample': sample,
                                'sample_alias': '{0}:sample:{1}'.format(
                                    study_acc, sample.name),
                                'run_alias': '{0}:assay:{1}'.format(
                                    study_acc, assay_seq_process.name),
                                'exp_alias': '{0}:generic_assay:{1}'.format(
                                    study_acc, '{0}:{1}'.format(
                                        iassay.filename[:-4],
                                        assay_seq_process.name)),
                                'data_files': []
                            }
                        datafiles = list(filter(
                            lambda datafile: isinstance(datafile, DataFile),
                            assay_seq_process.outputs
                        ))
                        for datafile in datafiles:
                            checksum = '00000000000000000000000000000000'
                            if datafilehashes is not None:
                                checksum = datafilehashes[datafile.filename]
                                # raises AttributeError if file not found
                            dot_indicies = [i for i, x in
                                            enumerate(datafile.filename) if
                                            x == '.']
                            file_ext = datafile.filename[dot_indicies[-1] + 1:]
                            if file_ext in ('.gz'):
                                # if is compressed, look for the actual ftype
                                try:
                                    filetype = datafile.filename[
                                        dot_indicies[-2]
                                        + 1:dot_indicies[-1]]
                                except IndexError:
                                    log.warning(
                                        "Could not infer SRA filetype for "
                                        "data file {filename}; defaulting to "
                                        "'other'".format(
                                            filename=datafile.filename))
                                    filetype = 'other'
                            else:
                                filetype = file_ext
                            assay_to_export['data_files'].append(
                                {
                                    'filename': datafile.filename,
                                    'filetype': filetype,
                                    'checksum': checksum
                                }
                            )
                        source = None
                        matching_sources = [
                            p.inputs for p in istudy.process_sequence
                            if sample in p.outputs]
                        if len(matching_sources[0]) == 1:
                            source = matching_sources[0][0]
                        assay_to_export['source'] = {
                            'name': source.name,
                            'characteristics': source.characteristics,
                        }
                        organism_charac = [
                            c for c in source.characteristics
                            if c.category.term == 'organism'][-1]
                        assay_to_export['source']['taxon_id'] = \
                            organism_charac.value.term_accession[
                            organism_charac.value.term_accession.index(
                                '_') + 1:]
                        assay_to_export['source']['scientific_name'] = \
                            organism_charac.value.term
                        curr_process = assay_seq_process
                        while curr_process.prev_process is not None:
                            assay_to_export[
                                curr_process.executes_protocol.protocol_type
                                .term] = curr_process
                            try:
                                curr_process = curr_process.prev_process
                            except AttributeError:
                                pass
                        target_taxon = get_pv(
                            assay_to_export['library construction'],
                            'target_taxon')
                        assay_to_export['target_taxon'] = target_taxon
                        assay_to_export['targeted_loci'] = False
                        assay_to_export['min_match'] = 0
                        # BEGIN genome seq library selection
                        if iassay.measurement_type.term in [
                            'genome sequencing',
                                'whole genome sequencing']:
                            library_source = get_pv(
                                assay_to_export['library construction'],
                                'library source')
                            if library_source.upper() not in [
                                'GENOMIC', 'GENOMIC SINGLE CELL',
                                    'METAGENOMIC', 'OTHER']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema {}'.format(
                                        library_source))
                                library_source = 'OTHER'

                            library_strategy = get_pv(
                                assay_to_export['library construction'],
                                'library strategy')
                            if library_strategy.upper() not in ['WGS',
                                                                'OTHER']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema {}'.format(
                                        library_strategy))
                                library_strategy = 'OTHER'

                            library_selection = get_pv(
                                assay_to_export['library construction'],
                                'library selection')
                            if library_selection not in ['RANDOM',
                                                         'UNSPECIFIED']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema '.format(
                                        library_selection))
                                library_selection = 'unspecified'

                            protocol = '\n protocol_description: {}'.format(
                                assay_to_export['library construction']
                                .executes_protocol.description)
                            mid_pv = get_pv(
                                assay_to_export['library construction'], 'mid')
                            if mid_pv is not None:
                                protocol += '\n mid: {}'.format(mid_pv.value)

                            assay_to_export['library_source'] = library_source
                            assay_to_export[
                                'library_strategy'] = library_strategy
                            assay_to_export[
                                'library_selection'] = library_selection
                            assay_to_export[
                                'library_construction_protocol'] = protocol

                            library_layout = get_pv(
                                assay_to_export['library construction'],
                                'library layout')
                            assay_to_export['library_layout'] = \
                                library_layout.lower()
                        # END genome seq library selection
                        # BEGIN environmental gene survey library selection
                        elif iassay.measurement_type.term in \
                                ['environmental gene survey']:
                            assay_to_export['library_source'] = 'METAGENOMIC'
                            assay_to_export['library_strategy'] = 'AMPLICON'
                            assay_to_export['library_selection'] = 'PCR'
                            library_layout = get_pv(
                                assay_to_export['library construction'],
                                'library layout')
                            assay_to_export['library_layout'] = \
                                library_layout.lower()
                            nucl_acid_amp = get_pv(
                                assay_to_export['library construction'],
                                'nucleic acid amplification')
                            if nucl_acid_amp is None:
                                nucl_acid_amp = get_pv(
                                    assay_to_export['library construction'],
                                    'nucl_acid_amp')

                            protocol = '\n protocol_description: '.format(
                                assay_to_export['library construction']
                                .executes_protocol.description)
                            mid_pv = get_pv(
                                assay_to_export['library construction'], 'mid')
                            if mid_pv is not None:
                                protocol += '\n mid: {}'.format(mid_pv)
                                assay_to_export['barcode'] = mid_pv
                                assay_to_export['min_match'] = len(mid_pv)
                            if nucl_acid_amp is not None:
                                protocol += '\n nucl_acid_amp: {}'\
                                    .format(nucl_acid_amp.value)
                            url = get_pv(
                                assay_to_export['library construction'], 'url')
                            if url is not None:
                                protocol += '\n url: '.format(
                                    nucl_acid_amp.value)
                            target_taxon = assay_to_export['target_taxon']
                            if target_taxon is not None:
                                protocol += '\n target_taxon: {}'.format(
                                    target_taxon)
                            target_gene = get_pv(
                                assay_to_export['library construction'],
                                'target_gene')
                            if target_gene is not None:
                                protocol += '\n target_gene: {}'.format(
                                    target_gene)
                            target_subfragment = get_pv(
                                assay_to_export['library construction'],
                                'target_subfragment')
                            if target_subfragment is not None:
                                protocol += '\n target_subfragment: {}'.format(
                                    target_subfragment)
                            pcr_primers = get_pv(
                                assay_to_export['library construction'],
                                'pcr_primers')
                            if pcr_primers is not None:
                                protocol += '\n pcr_primers: {}'.format(
                                    pcr_primers)
                            pcr_cond = get_pv(
                                assay_to_export['library construction'],
                                'pcr_cond')
                            if pcr_cond is not None:
                                protocol += '\n pcr_cond: {}'.format(pcr_cond)
                            assay_to_export[
                                'library_construction_protocol'] = protocol

                            if target_gene is not None:
                                assay_to_export['targeted_loci'] = True
                                assay_to_export['locus_name'] = target_gene
                        # END environmental gene survey library selection
                        # BEGIN metagenome seq library selection
                        elif iassay.measurement_type.term in \
                                ['metagenome sequencing']:
                            library_source = 'METAGENOMIC'
                            library_strategy = get_pv(
                                assay_to_export['library construction'],
                                'library strategy')
                            if library_strategy.upper() not in [
                                    'WGS', 'OTHER']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema '.format(
                                        library_strategy))
                                library_strategy = 'OTHER'

                            library_selection = get_pv(
                                assay_to_export['library construction'],
                                'library selection')
                            if library_selection not in \
                                    ['RANDOM', 'UNSPECIFIED']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema '.format(
                                        library_selection))
                                library_selection = 'unspecified'

                            protocol = '\n protocol_description: {}'.format(
                                assay_to_export['library construction']
                                .executes_protocol.description)
                            mid_pv = get_pv(
                                assay_to_export['library construction'], 'mid')
                            if mid_pv is not None:
                                protocol += '\n mid: {}'.format(mid_pv.value)

                            assay_to_export['library_source'] = library_source
                            assay_to_export['library_strategy'] = \
                                library_strategy
                            assay_to_export['library_selection'] = \
                                library_selection
                            assay_to_export[
                                'library_construction_protocol'] = protocol

                            library_layout = get_pv(
                                assay_to_export['library construction'],
                                'library layout')
                            assay_to_export['library_layout'] = \
                                library_layout.lower()
                        # END metagenome seq library selection
                        # BEGIN transciption profiling library selection
                        elif iassay.measurement_type.term in \
                                ['transcription profiling']:
                            library_source = get_pv(
                                assay_to_export['library construction'],
                                'library source')
                            if library_source is None:
                                # if not specified, select TRANSCRIPTOMIC by
                                # default
                                library_source = 'TRANSCRIPTOMIC'

                            if library_source.upper() not in \
                                    ['TRANSCRIPTOMIC',
                                     'TRANSCRIPTOMIC SINGLE CELL',
                                     'METATRANSCRIPTOMIC',
                                     'OTHER']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema {}'.format(
                                        library_source))
                                library_source = 'OTHER'

                            library_strategy = get_pv(
                                assay_to_export['library construction'],
                                'library strategy')
                            if library_strategy not in \
                                    ['RNA-Seq', 'ssRNA-Seq', 'miRNA-Seq',
                                     'ncRNA-Seq', 'FL-cDNA', 'EST', 'OTHER']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema {}'.format(
                                        library_strategy))
                                library_strategy = 'OTHER'

                            library_selection = get_pv(
                                assay_to_export['library construction'],
                                'library selection')
                            if library_selection not in \
                                    ['RT-PCR', 'cDNA', 'cDNA_randomPriming',
                                     'cDNA_oligo_dT', 'PolyA', 'Oligo-dT',
                                     'Inverse rRNA', 'Inverse rRNA selection',
                                     'CAGE', 'RACE', 'other']:
                                log.warning(
                                    'ERROR:value supplied is not compatible '
                                    'with SRA1.5 schema {}'.format(
                                        library_selection))
                                library_selection = 'other'

                            protocol = '\n protocol_description: {}'.format(
                                assay_to_export['library construction']
                                .executes_protocol.description)
                            assay_to_export['library_source'] = library_source
                            assay_to_export['library_strategy'] = \
                                library_strategy
                            assay_to_export['library_selection'] = \
                                library_selection
                            assay_to_export[
                                'library_construction_protocol'] = protocol

                            library_layout = get_pv(
                                assay_to_export['library construction'],
                                'library layout')
                            assay_to_export['library_layout'] = \
                                library_layout.lower()
                        # END transciption profiling library selection
                        else:
                            log.error(
                                'ERROR:Unsupported measurement type: {}'
                                .format(iassay.measurement_type.term))
                        mid_pv = get_pv(
                            assay_to_export['library construction'], 'mid')
                        assay_to_export['poolingstrategy'] = mid_pv
                        seq_instrument = get_pv(
                            assay_to_export['nucleic acid sequencing'],
                            'sequencing instrument')
                        assay_to_export['platform'] = seq_instrument
                        assays_to_export.append(assay_to_export)
            else:
                log.error(
                    'ERROR:Unsupported measurement/technology type {0}/{1}, '
                    'skipping assays'.format(
                        iassay.measurement_type.term,
                        iassay.technology_type.term))

        xexp_set_template = env.get_template('experiment_set.xml')
        xexp_set = xexp_set_template.render(assays_to_export=assays_to_export,
                                            study=istudy,
                                            sra_center_name=sra_center_name,
                                            sra_broker_name=sra_broker_name)
        xrun_set_template = env.get_template('run_set.xml')
        xrun_set = xrun_set_template.render(assays_to_export=assays_to_export,
                                            study=istudy,
                                            sra_center_name=sra_center_name,
                                            sra_broker_name=sra_broker_name)
        samples_to_export = list()
        for assay_to_export in assays_to_export:
            if len([s for s in samples_to_export if
                    s['sample_alias'] == assay_to_export['sample_alias']]) > 0:
                pass
            else:
                samples_to_export.append(assay_to_export)
        xsample_set_template = env.get_template('sample_set.xml')
        xsample_set = xsample_set_template.render(
            assays_to_export=samples_to_export, study=istudy,
            sra_center_name=sra_center_name, sra_broker_name=sra_broker_name)
        log.debug("SRA exporter: writing SRA XML files for study " + study_acc)

        # blitz out whitespaces with etree and format nicely with minidom
        def prettify(xmlstr):
            p = etree.XMLParser(remove_blank_text=True)
            exsub = etree.XML(xmlstr, parser=p)
            x = xml.dom.minidom.parseString(etree.tostring(exsub))
            return x.toprettyxml()

        # validate a doc against one of our SRA schemas
        def validate(docpath, schemaname):
            with open(os.path.join(os.path.dirname(__file__), 'resources',
                                   'sra_schemas', schemaname)) as xsd:
                doc = etree.parse(xsd)
                try:
                    schema = etree.XMLSchema(doc)
                    with open(docpath, 'r') as xsub_file:
                        doc = etree.parse(xsub_file)
                        try:
                            schema.assertValid(doc)
                        except etree.DocumentInvalid as e:
                            log.error(
                                'Schema validation failed on {}'
                                .format('{0}:\n{1}'.format(
                                        docpath, str(e))))
                except etree.XMLSchemaParseError as e:
                    log.error(e)

        if os.path.exists(export_path):
            with open(os.path.join(
                    export_path, 'submission.xml'), 'w') as xsub_file:
                print(prettify(xsub), file=xsub_file)
            validate(os.path.join(
                export_path, 'submission.xml'), 'SRA.submission.xsd')
            with open(os.path.join(
                    export_path, 'project_set.xml'), 'w') as xproj_set_file:
                print(prettify(xproj), file=xproj_set_file)
            validate(os.path.join(
                export_path, 'project_set.xml'), 'ENA.project.xsd')
            with open(os.path.join(
                    export_path, 'experiment_set.xml'), 'w') as xexp_set_file:
                print(prettify(xexp_set), file=xexp_set_file)
            validate(os.path.join(
                export_path, 'experiment_set.xml'), 'SRA.experiment.xsd')
            with open(os.path.join(
                    export_path, 'run_set.xml'), 'w') as xrun_set_file:
                print(prettify(xrun_set), file=xrun_set_file)
            validate(os.path.join(export_path, 'run_set.xml'), 'SRA.run.xsd')
            with open(os.path.join(
                    export_path, 'sample_set.xml'), 'w') as xsample_set_file:
                print(prettify(xsample_set), file=xsample_set_file)
            validate(os.path.join(
                export_path, 'sample_set.xml'), 'SRA.sample.xsd')
        else:
            raise NotADirectoryError(
                "export path '{}' is not a directory".format(export_path))


def create_datafile_hashes(fileroot, filenames):
    """
    Create md5 file dict for files in a directory with a particular extension

    :param fileroot: Root to directory containing files (assumes all in
    same dir)
    :param filenames: List of filenames of files to md5, assumed in fileroot
    :return: dict containing filenames and md5s

    Usage:
   / >>> filenames = [f for f in listdir('/path/') if f.endswith('.fastq.gz')]
  /  >>> create_datafile_hashes(fileroot='/path/s', filenames=filesnames)
    {
        'myfile1.gz': 'd41d8cd98f00b204e9800998ecf8427e',
        'myfile2.gz': 'd41d8cd98f00b204e9800998ecf8427e'
    }
    """
    def md5sum(filename):
        with open(filename, mode='rb') as f:
            d = hashlib.md5()
            for buf in iter(partial(f.read, 128), b''):
                d.update(buf)
        return d.hexdigest()

    from os.path import isfile, join
    datafilehashes = dict()
    for file in filenames:
        if isfile(join(fileroot, file)):
            datafilehashes[file] = md5sum(filename=join(fileroot, file))
        else:
            raise FileNotFoundError(
                '{} is not a file'.format(join(fileroot, file)))
    return datafilehashes



# -*- coding: utf-8 -*
"""Convert SampleTab to ISA-JSON"""
import json
import logging

from isatools import sampletab
from isatools.isajson import ISAJSONEncoder


logger = logging.getLogger('isatools')


def convert(source_sampletab_fp, target_json_fp):
    """ Converter for ISA-JSON to SampleTab.
    :param source_sampletab_fp: File descriptor of input SampleTab file
    :param target_json_fp: File descriptor to write output ISA JSON (must be
                           writeable)
    """
    ISA = sampletab.load(source_sampletab_fp)
    json.dump(ISA, fp=target_json_fp, cls=ISAJSONEncoder)



from collections import OrderedDict

from isatools.create.constants import SAMPLE, EXTRACT, LABELED_EXTRACT, DATA_FILE, ORGANISM_PART
from isatools.model import OntologyAnnotation

sample_list = [
        {
            'node_type': SAMPLE,
            'characteristics_category': OntologyAnnotation(term=ORGANISM_PART),
            'characteristics_value': 'liver',
            'size': 1,
            'technical_replicates': None,
            'is_input_to_next_protocols': True
        },
        {
            'node_type': SAMPLE,
            'characteristics_category': OntologyAnnotation(term=ORGANISM_PART),
            'characteristics_value': 'blood',
            'size': 5,
            'technical_replicates': None,
            'is_input_to_next_protocols': True
        },
        {
            'node_type': SAMPLE,
            'characteristics_category': OntologyAnnotation(term=ORGANISM_PART),
            'characteristics_value': 'heart',
            'size': 1,
            'technical_replicates': None,
            'is_input_to_next_protocols': True
        }
]

ms_assay_dict = OrderedDict([
    ('measurement_type', OntologyAnnotation(term='metabolite profiling')),
    ('technology_type', OntologyAnnotation(term='mass spectrometry')),
    ('extraction', {}),
    ('extract', [
        {
            'node_type': EXTRACT,
            'characteristics_category': OntologyAnnotation(term='extract type'),
            'characteristics_value': 'polar fraction',
            'size': 1,
            'is_input_to_next_protocols': True
        },
        {
            'node_type': EXTRACT,
            'characteristics_category': OntologyAnnotation(term='extract type'),
            'characteristics_value': 'lipids',
            'size': 1,
            'is_input_to_next_protocols': True
        }
    ]),
    ('labelling', {
        '#replicates': 2
    }),
    ('labelled extract', [
        {
            'node_type': LABELED_EXTRACT,
            'characteristics_category': OntologyAnnotation(term='labelled extract type'),
            'characteristics_value': '',
            'size': 1,
            'is_input_to_next_protocols': True
        }
    ]),
    ('mass spectrometry', {
        '#replicates': 2,
        OntologyAnnotation(term='instrument'): ['Agilent QTQF 6510'],
        OntologyAnnotation(term='injection_mode'): ['FIA', 'LC'],
        OntologyAnnotation(term='acquisition_mode'): ['positive mode']
    }),
    ('raw spectral data file', [
        {
            'node_type': DATA_FILE,
            'size': 2,
            'is_input_to_next_protocols': False
        }
    ])
])

annotated_ms_assay_dict = OrderedDict([
    ('measurement_type', OntologyAnnotation(term='metabolite profiling',
                                            term_accession='http://purl.obolibrary.org/obo/OBI_0000366')),
    ('technology_type', OntologyAnnotation(term='mass spectrometry',
                                           term_accession='http://purl.obolibrary.org/obo/OBI_0000470')),
    (OntologyAnnotation(
        term='extraction',
        term_accession='http://purl.obolibrary.org/obo/OBI_0302884'
    ), {}),
    (OntologyAnnotation(
        term='extract',
        term_accession='http://purl.obolibrary.org/obo/OBI_0000423'
    ), [
        {
            'node_type': EXTRACT,
            'characteristics_category': OntologyAnnotation(
                term='extract type',
                term_accession='http://purl.obolibrary.org/obo/NCIT_C82948'
            ),
            'characteristics_value': OntologyAnnotation(term='polar fraction'),
            'size': 1,
            'is_input_to_next_protocols': True
        },
        {
            'node_type': EXTRACT,
            'characteristics_category': OntologyAnnotation(
                term='extract type',
                term_accession='http://purl.obolibrary.org/obo/NCIT_C82948'
            ),
            'characteristics_value': OntologyAnnotation(term='lipids'),
            'size': 1,
            'is_input_to_next_protocols': True
        }
    ]),
    (OntologyAnnotation(
        term='labelling',
        term_accession='http://purl.obolibrary.org/obo/CHMO_0001675'
    ), {
        '#replicates': 2
    }),
    (OntologyAnnotation(
        term='labelled extract',
        term_accession='http://purl.obolibrary.org/obo/OBI_0000924'
    ), [
        {
            'node_type': LABELED_EXTRACT,
            'characteristics_category': OntologyAnnotation(
                term='labelled extract type',
                term_accession='http://purl.obolibrary.org/obo/NCIT_C43386'
            ),
            'characteristics_value': OntologyAnnotation(term=''),
            'size': 1,
            'is_input_to_next_protocols': True
        }
    ]),
    (OntologyAnnotation(
        term='mass spectrometry',
        term_accession='http://purl.obolibrary.org/obo/OBI_0200085'
    ), {
        '#replicates': 2,
        OntologyAnnotation(term='instrument'): [OntologyAnnotation(
            term='Agilent QTQF 6510',
            term_accession='http://purl.obolibrary.org/obo/MS_1000676'
        )],
        OntologyAnnotation(term='injection_mode'): [
            OntologyAnnotation(
                term='FIA',
                term_accession='http://purl.obolibrary.org/obo/MS_1000058'
            ),
            OntologyAnnotation(
                term='LC',
                term_accession=''
            )
        ],
        OntologyAnnotation(term='acquisition_mode'): [
            OntologyAnnotation(
                term='positive mode',
                term_accession='http://purl.obolibrary.org/obo/MS_1002807'
            )
        ]
    }),
    (OntologyAnnotation(
        term='raw spectral data file',
        term_accession='http://purl.obolibrary.org/obo/MS_1003083'
    ), [
        {
            'node_type': DATA_FILE,
            'size': 2,
            'is_input_to_next_protocols': False
        }
    ])
])

phti_assay_dict = OrderedDict([
    ('measurement_type', 'phenotyping'),
    ('technology_type', 'high-throughput imaging'),
            ('extraction', {}),
            ('extract', [
                {
                    'node_type': EXTRACT,
                    'characteristics_category': 'extract type',
                    'characteristics_value': 'supernatant',
                    'size': 1,
                    'technical_replicates': None,
                    'is_input_to_next_protocols': True
                },
                {
                    'node_type': EXTRACT,
                    'characteristics_category': 'extract type',
                    'characteristics_value': 'pellet',
                    'size': 1,
                    'technical_replicates': None,
                    'is_input_to_next_protocols': True
                }
            ]),
            ('phenotyping by high throughput imaging', {
                'instrument': ['lemnatech gigant'],
                'acquisition_mode': ['UV light', 'near-IR light', 'far-IR light', 'visible light'],
                'camera position': ['top','120 degree','240 degree','360 degree'],
                'imaging daily schedule': ['06.00','19.00']
            }),
            ('raw_spectral_data_file', [
                {
                    'node_type': DATA_FILE,
                    'size': 1,
                    'technical_replicates': 2,
                    'is_input_to_next_protocols': False
                }
            ])
        ])

lcdad_assay_dict = OrderedDict([
    ('measurement_type', 'metabolite identification'),
    ('technology_type', 'liquid chromatography diode-array detector'),
            ('extraction', {}),
            ('extract', [
                {
                    'node_type': EXTRACT,
                    'characteristics_category': 'extract type',
                    'characteristics_value': 'supernatant',
                    'size': 1,
                    'technical_replicates': None,
                    'is_input_to_next_protocols': True
                },
                {
                    'node_type': EXTRACT,
                    'characteristics_category': 'extract type',
                    'characteristics_value': 'pellet',
                    'size': 1,
                    'technical_replicates': None,
                    'is_input_to_next_protocols': True
                }
            ]),
            ('lcdad_spectroscopy', {
                'instrument': ['Shimadzu DAD 400'],
            }),
            ('raw_spectral_data_file', [
                {
                    'node_type': DATA_FILE,
                    'size': 1,
                    'technical_replicates': 2,
                    'is_input_to_next_protocols': False
                }
            ])
        ])

nmr_assay_dict = OrderedDict([
    ('measurement_type', OntologyAnnotation(term='metabolite profiling')),
    ('technology_type', OntologyAnnotation(term='nmr spectroscopy')),
    ('extraction', {}),
    ('extract', [
        {
            'node_type': EXTRACT,
            'characteristics_category': OntologyAnnotation(term='extract type'),
            'characteristics_value': 'supernatant',
            'size': 1,
            'is_input_to_next_protocols': True
        },
        {
            'node_type': EXTRACT,
            'characteristics_category': OntologyAnnotation(term='extract type'),
            'characteristics_value': 'pellet',
            'size': 1,
            'is_input_to_next_protocols': True
        }
    ]),
    ('nmr spectroscopy', {
        '#replicates': 2,
        OntologyAnnotation(term='instrument'): ['Bruker AvanceII 1 GHz'],
        OntologyAnnotation(term='acquisition_mode'): ['1D 13C NMR', '2D 13C-13C NMR'],
        OntologyAnnotation(term='pulse_sequence'): ['CPMG', 'watergate']
        # 'acquisition_mode': ['1D 13C NMR', '1D 1H NMR', '2D 13C-13C NMR'],
        # 'pulse_sequence': ['CPMG', 'TOCSY', 'HOESY', 'watergate']
    }),
    ('raw spectral data file', [
        {
            'node_type': DATA_FILE,
            'size': 1,
            'is_input_to_next_protocols': False
        }
    ])
])



# -*- coding: utf-8 -*-
"""Functions for importing from BioCrates"""
from time import time
import os
import pandas as pd
import glob
import logging


import subprocess
import sys
import uuid

import fileinput

from collections import defaultdict
from io import BytesIO
from shutil import rmtree
from zipfile import ZipFile

import bs4
from bs4 import BeautifulSoup


sys.modules['BeautifulSoup'] = bs4

__author__ = ['philippe.rocca-serra@oerc.ox.ac.uk',
              'massi@oerc.ox.ac.uk',
              'alfie']


DEFAULT_SAXON_EXECUTABLE = os.path.join(
    os.path.dirname(
        os.path.abspath(__file__)), 'resources', 'saxon9', 'saxon9he.jar')

BIOCRATES_DIR = os.path.join(os.path.dirname(__file__), 'resources',
                             'biocrates')

BIOCRATES_META_XSL_FILE = os.path.join(
    BIOCRATES_DIR, 'ISA-Team-Biocrates2ISATAB-refactor.xsl')

BIOCRATES_DATA_XSL_FILE = os.path.join(
    BIOCRATES_DIR, 'ISA-Team-Biocrates2MAF.xsl')

DESTINATION_DIR = 'output/isatab/'

SAMPLE_METADATA_INPUT_DIR = 'resources/biocrates/input-test/'

logger = logging.getLogger('isatools')


def replaceAll(file,searchExp,replaceExp):
    for line in fileinput.input(file, inplace=1):
        if searchExp in line:
            line = line.replace(searchExp,replaceExp)
        sys.stdout.write(line)


def zipdir(path, zip_file):
    """utility function to zip a whole directory"""
    # zip_file is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in files:
            zip_file.write(os.path.join(root, file))


def merge_biocrates_files(input_dir):
    """ --Biocrates2ISA support script:
    a simple script to merge several Biocrates xml files into one.
    Invoke this method before running a conversion from Biocrates to ISA-Tab
    if and only if the Biocrates study comes in more than one file why is
    this necessary? Biocrates export function limits the number of plates
    meaning that several xml files may be necessary to represent a complete
    study.
    input: a bag of Biocrates XML documents . Note: those should be from the
    same xsd.
    output: a rather large XML document compliant with Biocrates XSD. This
    file should be used as input to Biocrates2ISA xsl stylesheet.
    Important: Remember to up the memory available to python """

    contacts = []
    samples = []
    plate_set = []
    projects = []
    metabolites = []

    for i in glob.iglob(os.path.join(input_dir, '*.xml')):

        # f = open('/Users/Philippe/Documents/git/biocrates/Biocrates-TUM/
        # input-Biocrates-XML-files/'
        #          'all-biocrates-xml-files/'+i)

        # note the "xml" argument: this is to ensure that BeautifulSoup does
        # not lowercase attribute elements
        # (without, the resulting xml trips the xsl)
        # soup = BeautifulSoup(open('/Users/Philippe/Documents/git/
        # biocrates-DATA/Biocrates-TUM/input-Biocrates-XML-files/
        # all-biocrates-xml-files/'+i), "xml")

        soup = BeautifulSoup(open(i), "xml")

        contacts = contacts + soup.find_all('contact')
        samples = samples + soup.find_all('sample')
        projects = projects + soup.find_all('project')
        plate_set = plate_set + soup.find_all('plate')
        metabolites = metabolites + soup.find_all('metabolite')

    # fh = open("/Users/Philippe/Documents/git/biocrates-DATA/Biocrates-TUM/
    # biocrates-merged-output.xml",  'w+')
    fh = open("biocrates-merged-output.xml", 'w+')
    fh.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>")
    fh.write("<data xmlns=\"http://www.biocrates.com/metstat/result/xml_1.0\" "
             "swVersion=\"MetIQ_5.4.8-DB100-Boron-2607"
             "-DB100-2607\" "
             "concentrationUnit=\"uM\" "
             "dateExport=\"2015-10-28T10:37:23.484+01:00\" user=\"labadmin\">")

    # the order in which these objects are written reflect the Biocrates XML
    # structure.
    # known issue: some elements (e.g. contacts may be duplicated but as the
    # objects differ from one attribute value, it is not possible to remove
    # them without further alignment heuristic

    metabolites = list(set(metabolites))
    for element in metabolites:
        fh.write(str(element))

    plate_set = list(set(plate_set))
    for element in plate_set:
        fh.write(str(element))

    projects = list(set(projects))
    for element in projects:
        fh.write(str(element))

    samples = list(set(samples))
    for element in samples:
        fh.write(str(element))

    contacts = list(set(contacts))
    for element in contacts:
        fh.write(str(element.encode('utf-8')))

    fh.write("</data>")
    fh.close()

    return fh


def biocrates_to_isatab_convert(biocrates_filename, saxon_jar_path=DEFAULT_SAXON_EXECUTABLE):
    """
    Given a directory containing one or more biocrates xml filename, the method
    convert it to ISA-tab using an XSL 2.0 transformation.
    If the directory contains more than one biocrates xml file corresponding to
    one same study, a method is called to merge the xml documents together
    prior to the transformation
    The XSLT is invoked using an executable script.

    Notice: this method depend on SAXON XSLT Processor

    Parameters:
         :param biocrates_filename (str or list) - if a string must contain
         only comma separated valid SRA accessionnumbers
         :param saxon_jar_path str - if provided, must be a valid path pointing
                                     to SAXON Java JAR file. Otherwise, the
                                     default Saxon HE JAR will be used, if
                                     installed

    # merged = merge_biocrates_files("/path/to/bag/of/biocrates-DATA/
    input-Biocrates-XML-files/")
    # biocrates_to_isatab_convert('biocrates-merged-output.xml',
    saxon_jar_path="/path/to/saxon9-he-processor/saxon9he.jar")

    Returns:zp
        :returns io.BytesIO if at least one of the SRA instances has been
        successfully converted
                 (NOTE: should I return StringIO instead?)
    """
    res = None
    dir_name = uuid.uuid4().hex
    buffer = BytesIO()

    destination_dir = os.path.abspath(dir_name)
    logger.debug('Destination dir is: ' + destination_dir)
    logger.info('Destination dir is: ' + destination_dir)

    if os.path.exists(destination_dir):
        logger.debug('Removing dir' + destination_dir)
        logger.debug('Removing dir' + destination_dir)
        rmtree(destination_dir)

    try:
        INPUT_FILE = os.path.join(BIOCRATES_DIR, biocrates_filename)

        res = subprocess.call(['java', '-jar', saxon_jar_path, INPUT_FILE,
                               BIOCRATES_META_XSL_FILE,
                               'biocrates_filename=' + biocrates_filename,
                               'outputdir=' + destination_dir])

        logger.info('Subprocess Saxon exited with code: %d', res)

    except subprocess.CalledProcessError as err:
        logger.error("isatools.convert.biocrates2isatab: "
                     "CalledProcessError caught ", err.returncode)

        logger.debug(err)

    with ZipFile(buffer, 'w') as zip_file:
        # use relative dir_name to avoid absolute path on file names
        zipdir(dir_name, zip_file)
        logger.debug("!", zip_file.namelist())

    # clean up the target directory after the ZIP file has been closed
    # rmtree(destination_dir)

    buffer.seek(0)
    return buffer


__author__ = 'alfie'


def generatePolarityAttrsDict(plate, polarity, myAttrs, myMetabolites, mydict):
    usedop = plate.get('usedop')
    platebarcode = plate.get('platebarcode')
    injection = plate.find_all('injection', {'polarity': polarity})
    if len(injection) > 0:
        for pi in injection:
            myAttrList = []
            myMetabolitesList = []
            for p in pi.find_all('measure'):
                myrdfname = p.find_parent('injection').get(
                    'rawdatafilename').split('.')[0]
                for attr, value in p.attrs.items():
                    if attr != 'metabolite':
                        mydict[p.get('metabolite') + '-' + myrdfname + '-'
                               + attr + '-' + polarity.lower() + '-' + usedop
                               + '-' + platebarcode] = value
                        if attr not in myAttrList:
                            myAttrList.append(attr)
                myMblite = p.get('metabolite')
                if myMblite not in myMetabolitesList:
                    myMetabolitesList.append(myMblite)
            # it is assume that the rawdatafilename is unique in each of the
            # plate grouping and polarity
            myAttrs[pi.get('rawdatafilename').split('.')[0]] = myAttrList
        myMetabolites[usedop + '-' + platebarcode + '-' + polarity.lower()] = \
            myMetabolitesList
    return myAttrs, mydict


def generateAttrsDict(plate):
    # using dictionaries of lists
    posAttrs = defaultdict(list)
    negAttrs = defaultdict(list)
    posMetabolites = defaultdict(list)
    negMetabolites = defaultdict(list)
    mydict = {}
    posAttrs, mydict = generatePolarityAttrsDict(
        plate, 'POSITIVE', posAttrs, posMetabolites, mydict)
    negAttrs, mydict = generatePolarityAttrsDict(
        plate, 'NEGATIVE', negAttrs, negMetabolites, mydict)
    return posAttrs, negAttrs, posMetabolites, negMetabolites, mydict


def writeOutToFile(plate, polarity, usedop, platebarcode, output_dir,
                   uniqueAttrs, uniqueMetaboliteIdentifiers, mydict):
    pos_injection = plate.find_all('injection', {'polarity': polarity})
    if len(pos_injection) > 0:
        filename = 'm_MTBLSXXX_' + usedop + '_' + platebarcode + '_' + polarity.lower() \
            + '_maf.txt'
        logger.debug("filename: ", filename)
        with open(os.path.join(output_dir, filename), 'w') as file_handler:
            # writing out the header
            file_handler.write('metabolite_identification')
            for ua in uniqueAttrs:
                for myattr in uniqueAttrs[ua]:
                    file_handler.write('\t' + ua + '[' + myattr + ']')
            # now the rest of the rows
            for myMetabolite in uniqueMetaboliteIdentifiers[
                    usedop + '-'
                    + platebarcode + '-' + polarity.lower()]:
                file_handler.write('\n' + myMetabolite)
                for ua in uniqueAttrs:
                    for myattr in uniqueAttrs[ua]:
                        mykey = myMetabolite + '-' + ua + '-' + myattr \
                            + '-' + polarity.lower() + '-' + usedop \
                            + '-' + platebarcode
                        if mykey in mydict:
                            file_handler.write('\t' + mydict[mykey])
                        else:
                            file_handler.write('\t')
        file_handler.close()
        complete_MAF(os.path.join(output_dir, filename))


def complete_MAF(maf_stub):

    # data = pd_modin.read_csv(maf_stub, sep='\t')
    data = pd.read_csv(maf_stub, sep='\t')

    data.insert(1, "database_identifier", "")
    data.insert(2, "chemical_formula", "")
    data.insert(3, "smiles", "")
    data.insert(4, "inchi", "")
    data.insert(5, "mass_to_charge", "")
    data.insert(6, "modifications", "")
    data.insert(7, "charge", "")
    data.insert(8, "retention_time", "")
    data.insert(9, "taxid", "")
    data.insert(10, "species", "")
    data.insert(11, "database", "")
    data.insert(12, "database_version", "")
    data.insert(13, "reliability", "")
    data.insert(14, "uri", "")
    data.insert(15, "search_engine", "")
    data.insert(16, "search_engine_score", "")

    data.to_csv(maf_stub, sep='\t', encoding='utf-8', index=False)


def add_sample_metadata(sample_info_file, input_study_file):

    S_STUDY_LOC = os.path.join(DESTINATION_DIR, input_study_file)
    logger.debug("study file location:", S_STUDY_LOC)

    # data = pd_modin.read_csv(S_STUDY_LOC, sep='\t')
    data = pd.read_csv(S_STUDY_LOC, sep='\t')
    logger.debug("study file:", data)

    SAMPLE_METADATA_LOC = os.path.join(SAMPLE_METADATA_INPUT_DIR, sample_info_file)
    logger.debug("sample metadata file location:", SAMPLE_METADATA_LOC)

    # sample_desc = pd_modin.read_csv(SAMPLE_METADATA_LOC)
    sample_desc = pd.read_csv(SAMPLE_METADATA_LOC)
    logger.debug("sample metadata: ", sample_desc)

    # data.join(sample_desc, on='Characteristics[barcode identifier]')

    # result = data.join(sample_desc, on='Characteristics[barcode identifier]')

    # result = pd_modin.merge(data, sample_desc, on='Characteristics[barcode identifier]', left_index=True, how='outer')
    result = pd.merge(data, sample_desc, on='Characteristics[barcode identifier]', left_index=True, how='outer')
    cols = result.columns.tolist()
    logger.debug(cols)

    result = result[['Source Name', 'Material Type', 'Characteristics[barcode identifier]', 'internal_ID', 'resolute_ID',
                     'Characteristics[Organism]', 'Term Source REF', 'Term Accession Number',
                     'cellLine', 'cellosaurusID', 'Characteristics[chemical compound]',
                     'Protocol REF_x', 'Date', 'Sample Name', 'Characteristics[Organism part]', 'Term Source REF.1',
                     'Term Accession Number.1',
                     'Factor value [cellNumber]', 'Factor value [replicate]', 'Factor value [extractionVolume, µl]'
                     ]]

    result['cellosaurusID'] = result['cellosaurusID'].str.replace('cellosaurus:CVCL_',' https://web.expasy.org/cellosaurus/CVCL_')

    result = result.rename(columns={'internal_ID': 'Characteristics[internal_ID]',
                                    'resolute_ID': 'Characteristics[resolute_ID]',
                                    'Characteristics[chemical compound]': 'Characteristics[qc element]',
                                    'cellLine': 'Characteristics[cell line]',
                                    'cellosaurusID': 'Term Accession Number',
                                    'Protocol REF_x': 'Protocol REF',
                                    'Characteristics[Organism part]':'Characteristics[material type]',
                                    'Material Type': 'Characteristics[specimen type]',
                                    'Factor value [cellNumber]': 'Factor Value[cell seeding density]',
                                    'Factor value [replicate]': 'Factor Value[replicate number]',
                                    'Factor value [extractionVolume, µl]': 'Factor Value[extraction volume]'
                                    })

    study_factor_names="Study Factor Name\"" + "\t" + "\"cell seeding density\"" + "\t" + "\"replicate number\"" + "\t" + "\"extraction volume"

    replaceAll(DESTINATION_DIR+"i_inv_biocrates.txt", "Study Factor Name", study_factor_names)

    result.insert(9, "Term Source REF.2", "cellosaurus")

    result.insert(21, "Unit", "µl")

    result = result.rename(columns={'Term Source REF.1': 'Term Source REF',
                                    'Term Source REF.2': 'Term Source REF',
                                    'Term Accession Number.1': 'Term Accession Number'
                                    })

    result.to_csv(S_STUDY_LOC, sep='\t', encoding='utf-8', index=False)


def parseSample(biocrates_filename):

    folder_name = 'output/isatab/'
    output_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                              folder_name)

    # create the output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # file = sys.argv[1]
    # file=biocrates_filename
    INPUT_FILE = os.path.join(BIOCRATES_DIR, biocrates_filename)
    # open and read up the file
    handler = open(INPUT_FILE).read()
    soup = BeautifulSoup(handler, features="lxml")

    # get all the plates
    plates = soup.find_all('plate')
    for plate in plates:
        usedop = plate.get('usedop')
        # logger.debug(usedop)
        platebarcode = plate.get('platebarcode')
        # extracting the the distinct column labels, metabolites,
        # and rawdatafilename collect the data into a dictionary
        posAttrs, negAttrs, posMetabolites, negMetabolites, mydict = \
            generateAttrsDict(plate)
        # and start creating the sample tab files
        writeOutToFile(plate, 'POSITIVE', usedop, platebarcode, output_dir,
                       posAttrs, posMetabolites, mydict)

        writeOutToFile(plate, 'NEGATIVE', usedop, platebarcode, output_dir,
                       negAttrs, negMetabolites, mydict)


if __name__ == "__main__":
    start = time()
    biocrates_to_isatab_convert('biocrates-merged-output.xml', saxon_jar_path=DEFAULT_SAXON_EXECUTABLE)
    parseSample(biocrates_filename='biocrates-merged-output.xml')
    add_sample_metadata('EX0003_sample_metadata.csv', 's_study_biocrates.txt')
    end = time()
    logger.debug('The conversion took {:.2f} s.'.format(end - start))
# parseSample(sys.argv[1])
# uncomment to run test
# merged = merge_biocrates_files("/Users/Philippe/Documents/git/biocrates-DATA/Biocrates-TUM/input-Biocrates-XML-files/all-biocrates-xml-files/")

# Conc_R100028_export_incl_information_20200309.xml


# -*- coding: utf-8 -*-
"""The contents of this module are used solely for testing purposes in the
isatools test suite that is not packaged with the PyPI distribution"""
from __future__ import absolute_import
import logging
import os
import re
from os.path import basename

import datetime
import pandas as pd
from pandas.testing import assert_frame_equal

from isatools.isatab import read_investigation_file


log = logging.getLogger('isatools')


DATA_DIR = os.path.join(os.path.dirname(__file__), '..', '..', 'tests', 'data')

JSON_DATA_DIR = os.path.join(DATA_DIR, 'json')

UNIT_JSON_DATA_DIR = os.path.join(JSON_DATA_DIR, 'unit')

SRA_DATA_DIR = os.path.join(DATA_DIR, 'sra')

TAB_DATA_DIR = os.path.join(DATA_DIR, 'tab')

MAGETAB_DATA_DIR = os.path.join(DATA_DIR, 'magetab')

MZML_DATA_DIR = os.path.join(DATA_DIR, 'mzml')

SAMPLETAB_DATA_DIR = os.path.join(DATA_DIR, 'sampletab')

CONFIGS_DATA_DIR = os.path.join(DATA_DIR, 'configs')

SCHEMAS_DATA_DIR = os.path.join(DATA_DIR, 'schemas')

XML_CONFIGS_DATA_DIR = os.path.join(CONFIGS_DATA_DIR, 'xml')

DEFAULT2015_XML_CONFIGS_DATA_DIR = os.path.join(
    XML_CONFIGS_DATA_DIR, 'isaconfig-default_v2015-07-02')

SRA2016_XML_CONFIGS_DATA_DIR = os.path.join(
    XML_CONFIGS_DATA_DIR, 'isaconfig-seq_v2016-08-30-SRA1.5-august2014mod')

JSON_DEFAULT_CONFIGS_DATA_DIR = os.path.join(
    DATA_DIR, CONFIGS_DATA_DIR, 'json_default')

JSON_SRA_CONFIGS_DATA_DIR = \
    os.path.join(DATA_DIR, CONFIGS_DATA_DIR, 'json_sra')


_RX_CHARACTERISTICS = re.compile(r'Characteristics\[(.*?)\]')
_RX_PARAM_VALUE = re.compile(r'Parameter Value\[(.*?)\]')
_RX_FACTOR_VALUE = re.compile(r'Factor Value\[(.*?)\]')


def assert_tab_content_equal(fp_x, fp_y):
    """
    Test for equality of tab files, only down to level of content -
    should not be taken as canonical equality, but ather that all the expected
    content matches to both input files, but not the order in which they
    appear.

    For more precise equality, you will need to apply a configuration
        - use assert_tab_equal_by_config(fp_x, fp_y, config)
    :param fp_x: File descriptor of a ISAtab file
    :param fp_y: File descriptor of another  ISAtab file
    :return: True or False plus any AssertionErrors
    """

    def _assert_df_equal(x, y):
        # need to sort values to loosen up how equality is calculated
        try:
            assert_frame_equal(
                x.sort_values(by=x.columns[0]), y.sort_values(by=y.columns[0]))
            return True
        except AssertionError as e:
            lbl = datetime.datetime.now()
            x.to_csv('~/Downloads/test-isa-for-release/expected-{}.csv'.format(lbl))
            y.to_csv('~/Downloads/test-isa-for-release/actual-{}.csv'.format(lbl))
            log.error('Error thrown comparing two dataframes: {}'.format(e))
            log.error('x columns are: {}'.format(x.columns))
            log.error('y columns are: {}'.format(y.columns))
            log.error('x data are: {}'.format(x.to_numpy()))
            log.error('y data are: {}'.format(y.to_numpy()))
            return False

    if basename(fp_x.name).startswith('i_'):
        df_dict_x = read_investigation_file(fp_x)
        df_dict_y = read_investigation_file(fp_y)
        eq = True
        for k in df_dict_x.keys():
            dfx = df_dict_x[k]
            dfy = df_dict_y[k]
            if not isinstance(dfx, list):
                if not _assert_df_equal(dfx, dfy):
                    eq = False
                    break
            else:
                try:
                    for x, y in zip(sorted(dfx), sorted(dfy)):
                        if not _assert_df_equal(x, y):
                            eq = False
                            break
                except ValueError as e:
                    log.error(e)
        return eq
    else:

        def diff(a, b):
            b = set(b)
            return [aa for aa in a if aa not in b]

        import numpy as np
        df_x = pd.read_csv(fp_x, sep='\t', encoding='utf-8')
        df_y = pd.read_csv(fp_y, sep='\t', encoding='utf-8')
        try:
            # drop empty columns
            df_x = df_x.replace('', np.nan)
            df_x = df_x.dropna(axis=1, how='all')
            df_x = df_x.replace(np.nan, '')
            df_y = df_y.replace('', np.nan)
            df_y = df_y.dropna(axis=1, how='all')
            df_y = df_y.replace(np.nan, '')

            is_cols_equal = set(
                [x.split('.', 1)[0] for x in df_x.columns]) == \
                set([x.split('.', 1)[0] for x in df_y.columns])
            if not is_cols_equal:
                log.debug('x: ' + str(df_x.columns))
                log.debug('y: ' + str(df_y.columns))
                log.debug(diff(df_x.columns, df_y.columns))
                raise AssertionError('Columns in x do not match those in y')

            # reindex to add contexts for duplicate named columns
            # (i.e. Term Accession Number, Unit, etc.)
            newcolsx = list()
            for col in df_x.columns:
                newcolsx.append(col)
            for i, col in enumerate(df_x.columns):
                if any(RX.match(col) for RX in (
                        _RX_CHARACTERISTICS, _RX_PARAM_VALUE,
                        _RX_FACTOR_VALUE)):
                    try:
                        if 'Unit' in df_x.columns[i + 1]:
                            newcolsx[i + 1] = col + '/Unit'
                            if 'Term Source REF' in df_x.columns[i + 2]:
                                newcolsx[i + 2] = col + '/Unit/Term Source REF'
                            if 'Term Accession Number' in df_x.columns[i + 3]:
                                newcolsx[i + 3] = col + \
                                    '/Unit/Term Accession Number'
                        elif 'Term Source REF' in df_x.columns[i + 1]:
                            newcolsx[i + 1] = col + '/Term Source REF'
                            if 'Term Accession Number' in df_x.columns[i + 2]:
                                newcolsx[i + 2] = col + \
                                    '/Term Accession Number'
                    except IndexError:
                        pass
            df_x.columns = newcolsx
            newcolsy = list()
            for col in df_y.columns:
                newcolsy.append(col)
            for i, col in enumerate(df_y.columns):
                if any(RX.match(col) for RX in (
                        _RX_CHARACTERISTICS, _RX_PARAM_VALUE,
                        _RX_FACTOR_VALUE)):
                    try:
                        if 'Unit' in df_y.columns[i + 1]:
                            newcolsy[i + 1] = col + '/Unit'
                            if 'Term Source REF' in df_y.columns[i + 2]:
                                newcolsy[i + 2] = col + '/Unit/Term Source REF'
                            if 'Term Accession Number' in df_y.columns[i + 3]:
                                newcolsy[i + 3] = col + \
                                    '/Unit/Term Accession Number'
                        elif 'Term Source REF' in df_y.columns[i + 1]:
                            newcolsy[i + 1] = col + '/Term Source REF'
                            if 'Term Accession Number' in df_y.columns[i + 2]:
                                newcolsy[i + 2] = col + \
                                    '/Term Accession Number'
                    except IndexError:
                        pass
            df_y.columns = newcolsy
            for colx in df_x.columns:
                for eachx, eachy in zip(df_x.sort_values(by=colx)[colx],
                                        df_y.sort_values(by=colx)[colx]):
                    if eachx != eachy:
                        log.debug(df_x[colx])
                        log.debug(df_y[colx])
                        raise AssertionError('Value: ' + str(eachx) +
                                             ', does not match: ' + str(eachy))
            return True
        except AssertionError as e:
            log.error(str(e))
            return False


def sortlistsj(J):
    """Sorts lists in a JSON. It sorts in-place.

    :param J: Some JSON
    :return: None
    """
    if isinstance(J, dict):
        for k in J.keys():
            sortlistsj(J[k])
    elif isinstance(J, list):
        for o in J:
            if isinstance(o, dict) or isinstance(o, list):
                sortlistsj(o)
        if len(J) > 1:
            J.sort(key=lambda i: str(i.values()))


def sortlistsx(X):
    return X


def assert_json_equal(jx, jy):
    """Checks if two JSONs are equal

    :param jx: First JSON for comparison
    :param jy: Second JSON for comparison
    :return: True is equal, False if not
    """
    import json
    jx = json.loads(json.dumps(jx, sort_keys=True))
    jy = json.loads(json.dumps(jy, sort_keys=True))
    sortlistsj(jx)
    sortlistsj(jy)
    if jx == jy:
        return True
    else:
        from deepdiff import DeepDiff
        log.debug('DeepDiff={}'.format(DeepDiff(jx, jy)))
        return False


def assert_xml_equal(x1, x2):
    """Checks if two XMLs are equal

    :param x1: First XML for comparison
    :param x2: Second XML for comparison
    :return: True is equal, False if not
    """
    # Only counts tags of x1 and x2 to check if the right number appear in each

    def collect_tags(X):
        foundtags = set()
        for node in X.iter():
            foundtags.add(node.tag)
        return foundtags

    x1tags = collect_tags(x1)
    x2tags = collect_tags(x2)
    if len(x1tags - x2tags) > 0 or len(x2tags - x1tags) > 0:
        log.debug("Collected tags don't match: ", x1tags, x2tags)
        return False
    else:
        for tag in x1tags:
            tagcount1 = x1.xpath('count(//{})'.format(tag))
            tagcount2 = x2.xpath('count(//{})'.format(tag))
            if tagcount1 != tagcount2:
                log.debug('Counts of {0} tag do not match {1}:{2}'
                          .format(tag, int(tagcount1), int(tagcount2)))
                return False
        return True


def strip_ids(J):
    """Strips out @ids from some ISA-JSON

    :param J: Some JSON
    :return: None
    """
    for k, v in J.items():
        if isinstance(v, dict):
            strip_ids(v)
        elif isinstance(v, list):
            for i in v:
                strip_ids(i)
        else:
            if k == '@id':
                J[k] = ''



# -*- coding: utf-8 -*
"""Convert ISA-Tab to SRA-XML"""
import json
import logging
import os
from io import BytesIO, StringIO
from zipfile import ZipFile

from isatools.convert import isatab2json, json2sra


log = logging.getLogger('isatools')


def zipdir(path, zip_file):
    """utility function to zip only SRA xmls from a whole directory"""
    # zip_file is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in [f for f in files if f in [
            'submission.xml', 'project_set.xml', 'run_set.xml',
                'experiment_set.xml', 'sample_set.xml']]:
            zip_file.write(os.path.join(root, file),
                           arcname=os.path.join(os.path.basename(root), file))


BASE_DIR = os.path.dirname(__file__)
default_config_dir = os.path.join(BASE_DIR, '..', 'config', 'xml')


def convert(source_path, dest_path, sra_settings=None, validate_first=True):
    log.info("Converting ISA-Tab to JSON for %s", source_path)
    isa_json = isatab2json.convert(source_path, validate_first=validate_first)
    log.debug("Writing JSON to memory file")
    isa_json_fp = StringIO(json.dumps(isa_json))
    isa_json_fp.name = "BII-S-3.json"
    log.info("Converting JSON to SRA, writing to %s", dest_path)
    log.info("Using SRA settings %s", sra_settings)
    json2sra.convert(isa_json_fp, dest_path, sra_settings=sra_settings,
                     validate_first=False)
    log.info("Conversion from ISA-Tab to SRA complete")
    buffer = BytesIO()
    if os.path.isdir(dest_path):
        log.info("Zipping SRA files")
        with ZipFile(buffer, 'w') as zip_file:
            zipdir(dest_path, zip_file)
            log.debug("Zipped %s", zip_file.namelist())
        buffer.seek(0)
        log.info("Returning zipped files as memory file")
        return buffer
